# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#

import os
import sys
import json
import re
import logging
from collections import OrderedDict
from sgtn_properties import Properties

import ssl
if hasattr(ssl, '_create_unverified_context'):  # for python 2.7
    ssl._create_default_https_context = ssl._create_unverified_context

PY_VER = sys.version_info.major
UTF8 = 'utf-8'

if PY_VER == 2:
    import urllib2 as httplib

    # Support utf8 text
    reload(sys)
    sys.setdefaultencoding(UTF8)
else:
    import urllib.request as httplib

KEY_RESULT = 'result'
KEY_HEADERS = 'headers'
KEY_ERROR = 'error'

LOG_TYPE_INFO = 'info'
LOG_TYPE_ERROR = 'error'

# below keys are in lower case
LOCALE_MAP = {
    'zh-hant': 'zh-Hant',
    'zh-tw': 'zh-Hant',
    'zh-hans': 'zh-Hans',
    'zh-cn': 'zh-Hans'
    }

class FileUtil:

    @classmethod
    def read_text_file(cls, file_name):
        if os.path.exists(file_name) and os.path.isfile(file_name):
            f = open(file_name, 'rb')
            file_data = f.read()
            f.close()

            try:
                file_data = file_data.decode(UTF8)
                return file_data
            except Exception as e:
                return None
        return None

    @classmethod
    def parse_json(cls, text):
        if text:
            try:
                dict_data = json.loads(text, object_pairs_hook = OrderedDict)
                return dict_data
            except Exception as e:
                return None
        return None

    @classmethod
    def parse_yaml(cls, text):
        if text:
            try:
                import yaml
                dict_data = yaml.load(text, Loader = yaml.FullLoader)
                return dict_data
            except Exception as e:
                return None
        return None

    @classmethod
    def parse_datatree(cls, text):
        if text:
            data = cls.parse_yaml(text)
            if data is None:
                data = cls.parse_json(text)
            return data
        return None

    @classmethod
    def read_json_file(cls, file_name):
        file_data = cls.read_text_file(file_name)
        return cls.parse_json(file_data)

    @classmethod
    def read_datatree(cls, file_name):
        file_data = cls.read_text_file(file_name)
        return cls.parse_datatree(file_data)

        return cls.read_json_file(file_name)

    @classmethod
    def save_json_file(cls, file_name, dict):
        dir = os.path.dirname(file_name)
        if not os.path.exists(dir):
            os.makedirs(dir)
        if PY_VER == 2:
            f = open(file_name, 'w')
        else:
            f = open(file_name, 'w', encoding='utf-8')
        text = json.dumps(dict, ensure_ascii = False, indent = 2)
        f.write(text)
        f.close()

    @classmethod
    def get_dir_info(cls, dir_name):
        dir_list = []
        file_list = []

        try:
            ls = os.listdir(dir_name)
        except:
            pass
        else:
            for fn in ls:
                temp = os.path.join(dir_name, fn)
                if (os.path.isdir(temp)):
                    dir_list.append(fn)
                else:
                    file_list.append(fn)
        return dir_list, file_list

class NetUtil:

    simulate_data = None
    record_data = None

    @classmethod
    def _get_data(cls, url, request_headers):
        if not cls.simulate_data:
            req = httplib.Request(url)
            if request_headers:
                for key in request_headers:
                    req.add_header(key, request_headers[key])
            res_data = httplib.urlopen(req)

            headers = {}
            for h in res_data.headers:
                headers[h.lower()] = res_data.headers[h].lower()

            result = res_data.read()
            text = result.decode(UTF8)

            if cls.record_data is not None:
                header_part = json.dumps(request_headers) if request_headers else request_headers
                key = '%s<<headers>>%s' % (ur, header_part) if header_part else url
                cls.record_data[key] = {'text': text, 'headers': headers}
            return text, headers
        else:
            header_part = json.dumps(request_headers) if request_headers else request_headers
            key = '%s<<headers>>%s' % (url, header_part) if header_part else url
            kept = cls.simulate_data.get(key)
            if kept:
                if 'code' in kept:
                    if kept['code'] == 304:
                        raise Exception('Error 304:')
                return kept['text'], kept['headers']
        return None, None

    @classmethod
    def http_get_text(cls, url):
        text = None
        try:
            text, _ = cls._get_data(url, None)
        except Exception as e:
            pass
        return text

    @classmethod
    def http_get(cls, url, request_headers):
        ret = {}
        code = 400
        try:
            text, headers = cls._get_data(url, request_headers)
            ret[KEY_RESULT] = json.loads(text, object_pairs_hook = OrderedDict)
            ret[KEY_HEADERS] = headers
            code = 200
        except Exception as e:
            err_msg = str(e)
            parts = re.split("Error ([0-9]*):", err_msg)
            if len(parts) > 1:
                code = int(parts[1])

            if code != 304:
                ret[KEY_ERROR] = 'HTTP ERROR: %s' % str(e)
        return code, ret

    @classmethod
    def get_etag_maxage(cls, headers):
        if headers is None:
            return None, None
        etag = headers.get('etag')
        text = headers.get('cache-control')
        if text is None:
            return etag, None

        parts = re.split("max\\-age[ ]*\\=[ ]*([0-9]*)[ ]*", text)
        if len(parts) < 2:
            return etag, None
        return etag, float(parts[1])


class SysUtil:

    @classmethod
    def init_logger(cls, log_file, log_name):
        handler = logging.FileHandler(log_file)
        formatter = logging.Formatter('%(asctime)s %(message)s')
        handler.setFormatter(formatter)

        logger = logging.getLogger(log_name)
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)

        cls.log(logger, '')
        cls.log(logger, '--- start --- python --- %s' % (sys.version.split('\n')[0]))
        return logger

    @classmethod
    def log(cls, logger, text, log_type = LOG_TYPE_INFO):
        if logger:
            if log_type == LOG_TYPE_INFO:
                logger.info(text)
                return
            elif log_type == LOG_TYPE_ERROR:
                logger.error(text)
                return
        print(text)

    @classmethod
    def get_fallback_locale(cls, locale):
        parts = re.split("[\-_]", locale)
        parts[0] = parts[0].lower()
        if len(parts) > 1:
            parts[1] = parts[1].upper()
        locale = '-'.join(parts)

        fallback = LOCALE_MAP.get(locale.lower())
        if fallback:
            return fallback

        return parts[0]




# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#

import sys
from collections import OrderedDict

if sys.version_info.major == 2:
    # Support utf8 text
    reload(sys)
    sys.setdefaultencoding('utf-8')

MAX_LINE_BUFFER = 1024


class LineReader:

    def __init__(self, inCharBuf):
        self.lineBuf = [None] * MAX_LINE_BUFFER
        self.inLimit = 0
        self.inOff = 0

        self.inCharBuf = inCharBuf
        if self.inCharBuf:
            self.inLimit = len(self.inCharBuf)


    def read_line(self):
        length = 0
        c = 0

        skipWhiteSpace = True
        isCommentLine = False
        isNewLine = True
        appendedLineBegin = False
        precedingBackslash = False
        skipLF = False

        while True:
            if self.inOff >= self.inLimit:
                if length == 0 or isCommentLine:
                    return -1
                if precedingBackslash:
                    length -= 1
                return length

            #The line below is equivalent to calling a ISO8859-1 decoder.
            c = self.inCharBuf[self.inOff]
            self.inOff += 1

            if skipLF:
                skipLF = False
                if c == '\n':
                    continue

            if skipWhiteSpace:
                if c == ' ' or c == '\t' or c == '\f':
                    continue

                if not appendedLineBegin and (c == '\r' or c == '\n'):
                    continue

                skipWhiteSpace = False
                appendedLineBegin = False

            if isNewLine:
                isNewLine = False
                if c == '#' or c == '!':
                    isCommentLine = True
                    continue

            if c != '\n' and c != '\r':
                self.lineBuf[length] = c
                length += 1
                if length == len(self.lineBuf):
                    buf = [None] * length
                    self.lineBuf.extend(buf)

                #flip the preceding backslash flag
                if c == '\\':
                    precedingBackslash = not precedingBackslash
                else:
                    precedingBackslash = False
            else:
                #reached EOL
                if isCommentLine or length == 0:
                    isCommentLine = False
                    isNewLine = True
                    skipWhiteSpace = True
                    length = 0
                    continue

                if self.inOff >= self.inLimit:
                    if precedingBackslash:
                        length -= 1
                    return length

                if precedingBackslash:
                    length -= 1
                    #skip the leading whitespace characters in following line
                    skipWhiteSpace = True
                    appendedLineBegin = True
                    precedingBackslash = False
                    if c == '\r':
                        skipLF = True
                else:
                    return length


class Properties:

    def __init__(self):
        self.kvTable = None

    def parse(self, text):
        self.kvTable = OrderedDict()

        reader = LineReader(text)
        self.load(reader)
        return self.kvTable

    def put(self, key, value):
        oldValue = self.kvTable.get(key)
        self.kvTable[key] = value
        return oldValue

    def load(self, lr):
        convtBuf = [None] * MAX_LINE_BUFFER

        while True:
            limit = lr.read_line()
            if limit < 0:
                break

            c = 0
            keyLen = 0
            valueStart = limit
            hasSep = False

            precedingBackslash = False
            while True:
                if keyLen >= limit:
                    break

                c = lr.lineBuf[keyLen]
                #need check if escaped.
                if (c == '=' or c == ':') and not precedingBackslash:
                    valueStart = keyLen + 1
                    hasSep = True
                    break
                elif (c == ' ' or c == '\t' or c == '\f') and not precedingBackslash:
                    valueStart = keyLen + 1
                    break

                if c == '\\':
                    precedingBackslash = not precedingBackslash
                else:
                    precedingBackslash = False

                keyLen += 1

            while True:
                if valueStart >= limit:
                    break

                c = lr.lineBuf[valueStart]
                if c != ' ' and c != '\t' and c != '\f':
                    if not hasSep and (c == '=' or c == ':'):
                        hasSep = True
                    else:
                        break

                valueStart += 1

            key = self.load_convert(lr.lineBuf, 0, keyLen, convtBuf)
            value = self.load_convert(lr.lineBuf, valueStart, limit-valueStart, convtBuf)
            self.put(key, value)

    def load_convert(self, inText, off, length, convtBuf):
        if len(convtBuf) < length:
            newLen = length * 2
            convtBuf = [None] * newLen

        outText = convtBuf
        outLen = 0
        end = off + length

        while True:
            if off >= end:
                break

            aChar = inText[off]
            off += 1
            if aChar == '\\':
                aChar = inText[off]
                off += 1
                if aChar == 'u':
                    #Read the xxxx
                    value = 0
                    for i in range(4):
                        aChar = inText[off]
                        off += 1

                        if aChar >= '0' and aChar <= '9':
                            value = (value << 4) + int(aChar-'0')
                        elif aChar >= 'a' and aChar <= 'f':
                            value = (value << 4) + 10 + int(aChar-'a')
                        elif aChar >= 'A' and aChar <= 'F':
                            value = (value << 4) + 10 + int(aChar-'A')
                        else:
                            return None

                    outText[outLen] = rune(value)
                    outLen += 1
                else:
                    if aChar == 't':
                        aChar = '\t'
                    elif aChar == 'r':
                        aChar = '\r'
                    elif aChar == 'n':
                        aChar = '\n'
                    elif aChar == 'f':
                        aChar = '\f'

                    outText[outLen] = aChar
                    outLen += 1
            else:
                outText[outLen] = aChar
                outLen += 1

        return ''.join(outText[:outLen])



# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#
import unittest

import os
import time
import json

import sys
sys.path.append('../sgtnclient')
import I18N
from sgtn_util import FileUtil, NetUtil

from util import Util, allTestData

plans = [
    {'product': 'PYTHON1', 'config': 'config/sgtn_online_only.yml'},
    {'product': 'PYTHON2', 'config': 'config/sgtn_offline_only.yml'},
    {'product': 'PYTHON3', 'config': 'config/sgtn_online_localsource.yml'},
    {'product': 'PYTHON4', 'config': 'config/sgtn_online_default.yml'},
    {'product': 'PYTHON5', 'config': 'config/sgtn_offline_default.yml'}
    ]
VERSION = '1.0.0'
COMPONENT = 'about'
LOCALE = 'de'


class TestClient(unittest.TestCase):

    def prepare_sub_path(self, sub):
        current_path = os.path.dirname(__file__)
        sub_path = os.path.join(current_path, sub)
        if not os.path.exists(sub_path):
            os.makedirs(sub_path)

    def show(self, text1, text2, value):
        print('--- %s --- %s --- %s ---' % (text1, text2, value))

    def check_locale(self, trans, locale):
        fallback_locale = trans.get_locale_supported(locale)
        self.show('locale', locale, fallback_locale)

    def dict2string(self, dict):
        return json.dumps(dict, ensure_ascii = False, indent = 2)

    def need_wait(self, cfg_info):
        if (cfg_info.get('local') and cfg_info.get('remote')):
            return True
        return False

    def do_test(self, plan):
        cfg = I18N.add_config_file(
            plan['config'],
            {'$PRODUCT': plan['product'], '$VERSION': VERSION})
        self.assertEqual(cfg.get_info()['version'], VERSION)

        start = time.time()
        I18N.set_current_locale(LOCALE)
        I18N.set_current_locale(LOCALE)
        current = I18N.get_current_locale()
        print('--- current --- %s ---' % current)
        self.assertEqual(current, 'de')

        rel = I18N.get_release(plan['product'], VERSION)

        cfg = rel.get_config()
        cfg_info = cfg.get_info()
        self.show('config', 'info', self.dict2string(cfg_info))

        trans = rel.get_translation()
        self.check_locale(trans, 'ZH_cn')
        self.check_locale(trans, 'EN_us')

        Util.run_test_data(self, trans, 'TestGetString1')
        Util.run_test_data(self, trans, 'TestGetString1T')
        if Util.is_async_supported():
            Util.run_test_data(self, trans, 'TestGetString1A')
        Util.run_test_data(self, trans, 'TestGetString2')

        groupName = 'TestGetStringSameLocale'
        if cfg.get_info()['default_locale'] != cfg.get_info()['source_locale']:
            groupName = 'TestGetStringDifferentLocale'
        Util.run_test_data(self, trans, groupName)
        Util.run_test_data(self, trans, 'TestGetStringTemp')

        I18N.set_current_locale(LOCALE)
        found = trans.get_string('TT', 'about.title')
        print('--- found --- wrong component --- %s ---' % found)

        if (self.need_wait(cfg_info)):
            time.sleep(5)

        found = trans.get_string(None, 'about.title', format_items = ['11', '22'])
        print('--- found --- format in array --- %s ---' % found)
        self.assertEqual(found, 'Über Version 22 of Product 11')
        found = trans.get_string(None, 'about.title2', format_items = {'x': 'ee', 'y': 'ff'})
        print('--- found --- format in dict --- %s ---' % found)
        self.assertEqual(found, 'Über Version ee of Product ff')

        spent = time.time() - start

        data = trans.get_locale_strings('en-US', True)
        print('--- source --- en-US --- %s ---' % data)

        data = trans.get_locale_strings('en-US', False)
        print('--- translate --- en-US --- %s ---' % data)

        data = trans.get_locale_strings('de', False)
        print('--- translate --- de --- %s ---' % data)

        if (self.need_wait(cfg_info)):
            Util.run_test_data(self, trans, 'TestGetString3')

        if NetUtil.record_data is not None:
            time.sleep(5)
            FileUtil.save_json_file('data/simulate.json', NetUtil.record_data)
        time.sleep(1)
        print('--- test --- end --- %s ---' % spent)

    def test_api(self):
        print('\n--- unittest --- %s --- python %s\n' % (
            sys._getframe().f_code.co_name, sys.version_info.major))

        NetUtil.simulate_data = Util.load_response(['data/http_response.txt'])
        #NetUtil.record_data = {}

        Util.load_test_data(['data/test_define.txt'])

        print('--- test start ---')

        self.prepare_sub_path('log')
        self.prepare_sub_path('singleton')

        for i in range(len(plans)):
            self.do_test(plans[i])


if __name__ == '__main__':
    unittest.main()



# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#

import sys
import unittest

sys.path.append('../sgtnclient')
if sys.version_info.major == 2:
    # Support utf8 text
    reload(sys)
    sys.setdefaultencoding('utf-8')

from sgtn_util import FileUtil, NetUtil, SysUtil

from util import Util


class TestClient(unittest.TestCase):

    def test_file_util(self):
        print('\n--- unittest --- %s --- python %s\n' % (
            sys._getframe().f_code.co_name, sys.version_info.major))

        text = FileUtil.read_text_file('data/data.txt')
        self.assertIn('cc=AA{x}BB{y}CC', text)
        
        dt = FileUtil.read_json_file('data/data.json')
        self.assertEqual(dt['aa'], 'aaa')
        print('--- json --- %s ---' % dt)

        dt['add'] = 'über'
        FileUtil.save_json_file('./log/data2.json', dt)
        dtLoad = FileUtil.read_json_file('./log/data2.json')
        self.assertEqual(dtLoad['add'], 'über')
        
        dt = FileUtil.read_datatree('config/sgtn_online_only.yml')
        self.assertEqual(dt['cache_type'], 'by_key')
        print('--- yaml --- %s ---' % dt['cache_type'])
        
        dir_list, file_list = FileUtil.get_dir_info('data')
        print('--- dir_list --- %s ---' % dir_list)
        print('--- file_list --- %s ---' % len(file_list))
        self.assertIn('http_response.txt', file_list)

    def test_net_util(self):
        print('\n--- unittest --- %s --- python %s\n' % (
            sys._getframe().f_code.co_name, sys.version_info.major))

        NetUtil.simulate_data = Util.load_response(['data/http_response.txt'])

        dt = FileUtil.read_datatree('config/sgtn_online_only.yml')
        online_url = dt['online_service_url']
        parts = online_url.split('/')[:3]
        parts.append('i18n/api/v2/translation/products/PYTHON1/versions/1.0.0/localelist')
        url = '/'.join(parts)
        
        text = NetUtil.http_get_text(url)
        self.assertIn('productName', text)
        
        code, dt = NetUtil.http_get(url, None)
        self.assertEqual(code, 200)
        self.assertIn('data', dt['result'])

        etag, max_age = NetUtil.get_etag_maxage(dt['headers'])
        headers = {'If-None-Match': etag}        
        code, dt = NetUtil.http_get(url, headers)
        self.assertEqual(code, 304)

    def test_sys_util(self):
        print('\n--- unittest --- %s --- python %s\n' % (
            sys._getframe().f_code.co_name, sys.version_info.major))

        locale = SysUtil.get_fallback_locale('ZH_cn')
        print('--- locale --- %s ---' % locale)
        self.assertEqual(locale, 'zh-Hans')

        locale = SysUtil.get_fallback_locale('EN_us')
        print('--- locale --- %s ---' % locale)
        self.assertEqual(locale, 'en')


if __name__ == '__main__':
    unittest.main()



# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#

import os
import sys

_libPath = os.path.dirname(__file__)
if _libPath not in sys.path:
    sys.path.append(_libPath)

from sgtn_util import FileUtil


_client_manager = None
def _get_client_manager():
    global _client_manager
    if _client_manager is None:
        from sgtn_client import SingletonClientManager
        _client_manager = SingletonClientManager()
    return _client_manager


class Config(object):
    """Config interface"""

    def get_config_data(self):
        raise Exception()

    def get_info(self):
        raise Exception(NOT_IMP_EXCEPTION)

class Translation(object):
    """Translation interface"""

    def get_string(self, component, key, **kwargs):
        raise Exception(NOT_IMP_EXCEPTION)

    def get_locale_strings(self, locale):
        raise Exception(NOT_IMP_EXCEPTION)

    def get_locale_supported(self, locale):
        raise Exception(NOT_IMP_EXCEPTION)

class Release(object):
    """Release interface"""

    def get_config(self):
        """get config interface Config"""
        raise Exception(NOT_IMP_EXCEPTION)

    def get_translation(self):
        """get translation interface Translation"""
        raise Exception(NOT_IMP_EXCEPTION)


def add_config_file(config_file, replaceMap = None):
    return _get_client_manager().add_config_file(config_file, replaceMap)

def add_config(base_path, config_data):
    return _get_client_manager().add_config(base_path, config_data)

def set_current_locale(locale):
    _get_client_manager().set_current_locale(locale)

def get_current_locale():
    """get string of locale"""
    return _get_client_manager().get_current_locale()

def get_release(product, version):
    """get release interface Release"""
    return _get_client_manager().get_release(product, version)




# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#

import re
import locale as locale_system


class SingletonLocale(object):

    def __init__(self, locale):
        self._localeList = [locale]

    def get_near_locale_list(self):
        return self._localeList

    def add_near_locale(self, locale):
        if locale in self._localeList:
            return False

        self._localeList.append(locale)
        return True

    def get_count(self):
        return len(self._localeList)

    def get_near_locale(self, index):
        if index < 0 or index >= self.get_count():
            return None
        return self._localeList[index]

    def get_original_locale(self):
        return self.get_near_locale(0)

    def compare(self, singletonLocale):
        if singletonLocale is None:
            return False

        return self.is_in_locale_list(singletonLocale.get_near_locale_list())

    def is_in_locale_list(self, checkList):
        if checkList is None:
            return False

        for i in range(self.get_count()):
            if self.get_near_locale(i) in checkList:
                return True

        return False

    def find_item(self, items, start):
        for i in range(start, self.get_count()):
            nearLocale = self.get_near_locale(i)
            item = items.get(nearLocale)
            if item:
                return item
        return None

    def set_items(self, items, item):
        for i in range(self.get_count()):
            nearLocale = self.get_near_locale(i)
            items[nearLocale] = item


class SingletonLocaleUtil(object):

    DEFAULT_LOCALE = "en-US"
    FALLBACK = {
      'zh-CN': 'zh-Hans',
      'zh-TW': 'zh-Hant',
      'zh-HANS': 'zh-Hans',
      'zh-HANT': 'zh-Hant'
    }

    LocaleFallbackMap = {}
    SystemLocale = None

    @classmethod
    def get_singleton_locale(cls, locale):
        if locale is None:
            return cls.get_singleton_locale(SingletonLocaleUtil.DEFAULT_LOCALE)

        singletonLocale = cls.LocaleFallbackMap.get(locale.lower())
        if singletonLocale:
            return singletonLocale

        parts = re.split('[\_|\-]', locale)
        parts[0] = parts[0].lower()
        if len(parts) > 1:
            parts[1] = parts[1].upper()
        original = '-'.join(parts)
        singletonLocale = SingletonLocale(original)
        fallback = cls.FALLBACK.get(original)
        if fallback:
            singletonLocale.add_near_locale(fallback)
        elif len(parts) > 1:
            singletonLocale.add_near_locale(parts[0])

        cls.LocaleFallbackMap[locale.lower()] = singletonLocale
        return singletonLocale



# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#

import sys
import unittest

sys.path.append('../sgtnclient')
if sys.version_info.major == 2:
    # Support utf8 text
    reload(sys)
    sys.setdefaultencoding('utf-8')

from sgtn_properties import Properties


class TestClient(unittest.TestCase):

    def test_properties_parser(self):
        print('\n--- unittest --- %s --- python %s\n' % (
            sys._getframe().f_code.co_name, sys.version_info.major))

        text = u'qq1=qqq\n#the middle one\nbb2:bbb #tail\ncc3 abc好abc好'
        print(text)
        
        p = Properties()
        m = p.parse(text)

        self.assertEqual(m['bb2'], 'bbb #tail')
        print('--- map --- %s ---' % m)

        import json
        t = json.dumps(m, indent=2, ensure_ascii=False)
        print(t)


if __name__ == '__main__':
    unittest.main()



# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#

from collections import OrderedDict
import threading
lock = threading.Lock()

from sgtn_locale import SingletonLocale, SingletonLocaleUtil

_indexLocaleItem = 0


class SingletonByKeyItem(object):

    def __init__(self, componentIndex, itemIndex):
        self._componentIndex = componentIndex

        self._pageIndex = itemIndex // SingletonByKey.PAGE_MAX_SIZE
        self._indexInPage = itemIndex % SingletonByKey.PAGE_MAX_SIZE

        self._sourceStatus = 0x01;
        self._next = None


class SingletonByKeyTable(object):

    def __init__(self, max):
        self._max = max
        self._table = [None] * max

    def get_page(self, id):
        return self._table[id]

    def new_page(self, id):
        array = [None] * self._max
        self._table[id] = array
        return array

    def get_item(self, pageIndex, indexInPage):
        array = self.get_page(pageIndex)
        if array is None:
            return None
        return array[indexInPage]

    def set_item(self, pageIndex, indexInPage, item):
        array = self.get_page(pageIndex)
        if array is None:
            array = self.new_page(pageIndex)

        array[indexInPage] = item

    def get_item_by_one_index(self, index):
        pageIndex = index // self._max
        indexInPage = index % self._max
        return self.get_item(pageIndex, indexInPage)

    def set_item_by_one_index(self, index, item):
        pageIndex = index // self._max
        indexInPage = index % self._max
        self.set_item(pageIndex, indexInPage, item)


class SingletonByKeyComponents(object):

    def __init__(self):
        self._count = 0
        self._componentTable = []
        self._componentIndexTable = {}

    def get_id(self, component):
        if not component:
            return -1

        componentIndex = self._componentIndexTable.get(component)
        if componentIndex != None:
            return componentIndex

        self._componentTable.append(component)
        self._componentIndexTable[component] = self._count
        self._count += 1
        return self._count - 1

    def get_name(self, id):
        if id < 0 or id >= self._count:
            return None

        return _componentTable[id]


class SingletonByKeyLocale(object):

    def __init__(self, bykey, locale, asSource):
        global _indexLocaleItem
        _indexLocaleItem += 1
        self._indexLocaleItem = _indexLocaleItem

        self._bykey = bykey
        self._locale = locale
        self._singletonLocale = SingletonLocaleUtil.get_singleton_locale(locale)
        self._asSource = asSource
        self._isSourceLocale = self._singletonLocale.compare(bykey._singletonLocaleSource)

        self._messages = SingletonByKeyTable(SingletonByKey.PAGE_MAX_SIZE)
        self._components = SingletonByKeyTable(SingletonByKey.COMPONENT_PAGE_MAX_SIZE)

    def check_task(self, componentIndex, needCheck):
        if componentIndex >= 0 and needCheck:
            componentObj = self._components.get_item_by_one_index(componentIndex)
            if componentObj != None and componentObj.task != None:
                componentObj.task.check()

    def get_message(self, componentIndex, pageIndex, indexInPage, needCheck = True):
        self.check_task(componentIndex, needCheck)
        return self._messages.get_item(pageIndex, indexInPage)

    def set_message(self, message, componentObject, componentIndex, pageIndex, indexInPage):
        if componentObject:
            self._components.set_item_by_one_index(componentIndex, componentObject)

        self._messages.set_item(pageIndex, indexInPage, message)
        return True;


class SingletonLookup(object):

    def __init__(self, key, componentIndex, message):
        self._key = key
        self._componentIndex = componentIndex
        self._message = message

        self._add = 0
        self._aboveItem = None
        self._currentItem = None


class SingletonByKey(object):

    PAGE_MAX_SIZE = 1024;
    COMPONENT_PAGE_MAX_SIZE = 128;

    def __init__(self, localeSource, localeDefault, isDifferent, cacheType):
        self._itemCount = 0
        self._keyAttrTable = {}
        self._items = SingletonByKeyTable(SingletonByKey.PAGE_MAX_SIZE)
        self._componentTable = SingletonByKeyComponents()

        self._singletonLocaleSource = SingletonLocaleUtil.get_singleton_locale(localeSource)

        self._sources = {}
        self._locales = {}
        self._onlyByKey = (cacheType == 'by_key')

        self._isDifferent = isDifferent
        self._sourceLocal = None
        self._sourceRemote = None

        self._defaultLocale = localeDefault
        self._defaultRemote = None

    def set_item(self, item, pageIndex, indexInPage):
        self._items.set_item(pageIndex, indexInPage, item)
        return True

    def get_and_add_itemcount(self):
        count = self._itemCount
        self._itemCount += 1
        return count

    def get_locale_item(self, locale, asSource):
        table = self._sources if asSource else self._locales
        item = table.get(locale)
        if item is None:
            singletonLocale = SingletonLocaleUtil.get_singleton_locale(locale)
            for oneLocale in table:
                oneSingletonLocale = SingletonLocaleUtil.get_singleton_locale(oneLocale)
                if singletonLocale.compare(oneSingletonLocale):
                    item = table[oneLocale]
                    break
            if item is None:
                item = SingletonByKeyLocale(self, locale, asSource)
            table[locale] = item
        return item

    def get_component_index(self, component):
        return self._componentTable.get_id(component)

    def get_string(self, key, componentIndex, localeItem, needFallback = False):
        if componentIndex < 0 and not self._onlyByKey:
            return None

        item = self._keyAttrTable.get(key)

        if componentIndex >= 0:
            while item:
                if item._componentIndex == componentIndex:
                    break
                item = item._next

        if item is None:
            localeItem.check_task(componentIndex, needFallback)
            return None

        if not needFallback:
            message = localeItem.get_message(componentIndex, item._pageIndex, item._indexInPage, False)
            return message

        message = None
        if item._sourceStatus & 0x01 == 0x01:
            message = localeItem.get_message(componentIndex, item._pageIndex, item._indexInPage)
            if message != None:
                return message

        if self._isDifferent:
            if not self._defaultRemote:
                self._defaultRemote = self.get_locale_item(self._defaultLocale, False)
            message = self._defaultRemote.get_message(componentIndex, item._pageIndex, item._indexInPage)

        if message is None:
            if item._sourceStatus & 0x04 == 0x04:
                message = self._sourceLocal.get_message(componentIndex, item._pageIndex, item._indexInPage)
            elif item._sourceStatus & 0x03 == 0x03:
                message = self._sourceRemote.get_message(componentIndex, item._pageIndex, item._indexInPage)

        if message is None:
            message = key
        return message

    def new_key_item(self, componentIndex):
        itemIndex = self.get_and_add_itemcount()
        item = SingletonByKeyItem(componentIndex, itemIndex)
        self.set_item(item, item._pageIndex, item._indexInPage)
        return item

    def _find_or_add(self, lookup):
        item = self._keyAttrTable.get(lookup._key)
        if item is None: # This is new
            lookup._currentItem = self.new_key_item(lookup._componentIndex)
            lookup._add = 1
            return

        while item:
            if item._componentIndex == lookup._componentIndex: # Found
                lookup._currentItem = item
                return
            lookup._aboveItem = item
            item = item._next

        lookup._currentItem = self.new_key_item(lookup._componentIndex)
        lookup._add = 2

    def do_set_string(self, key, componentObject, componentIndex, localeItem, message):
        lookup = SingletonLookup(key, componentIndex, message)
        self._find_or_add(lookup)

        item = lookup._currentItem
        if item is None:
            return False

        done = localeItem.set_message(message, componentObject, componentIndex, item._pageIndex, item._indexInPage)
        if done and localeItem._isSourceLocale:
            status = item._sourceStatus
            if localeItem._asSource:
                self._sourceLocal = localeItem;
                status |= 0x04
            elif localeItem._isSourceLocale:
                self._sourceRemote = localeItem
                status |= 0x02

            if (status & 0x06) != 0x06:
                status |= 0x01
            else:
                localSource = self._sourceLocal.get_message(componentIndex, item._pageIndex, item._indexInPage, False)
                remoteSource = self._sourceRemote.get_message(componentIndex, item._pageIndex, item._indexInPage, False)
                if localSource == remoteSource:
                    status |= 0x01
                else:
                    status &= 0x06
            item._sourceStatus = status

        # Finally, it's added in the table after it has been prepared to keep reading correct.
        if lookup._add == 1:
            self._keyAttrTable[key] = lookup._currentItem
        elif lookup._add == 2:
            lookup._aboveItem._next = lookup._currentItem
        return done;

    def set_string(self, key, componentObject, componentIndex, localeItem, message):
        if message is None or key is None or localeItem is None:
            return False

        text = self.get_string(key, componentIndex, localeItem)
        if message != text:
            with lock:
                text = self.get_string(key, componentIndex, localeItem)
                if message != text:
                    return self.do_set_string(key, componentObject, componentIndex, localeItem, message)
        return False

    def get_key_item(self, pageIndex, indexInPage):
        array = self._items.get_page(pageIndex)
        if array is None:
            return None
        return array[indexInPage]

    def get_messages(self, componentIndex, localeItem):
        messages = OrderedDict()
        if componentIndex >= 0 and localeItem:
            pages = {}
            for i in range(SingletonByKey.PAGE_MAX_SIZE):
                array = localeItem._messages.get_page(i)
                if array is None:
                    continue

                for k in range(SingletonByKey.PAGE_MAX_SIZE):
                    text = array[k]
                    if text != None:
                        item = self.get_key_item(i, k)
                        if item:
                            if i not in pages:
                                pages[i] = {}
                            pages[i][k] = ''

            for key in self._keyAttrTable:
                item = self._keyAttrTable.get(key)
                if item._pageIndex in pages:
                    array = pages[item._pageIndex]
                    if item._indexInPage in array:
                        messages[key] = self.get_string(key, componentIndex, localeItem)
        return messages



# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#

import sys
import os
import json
import re
import time
import copy
import logging
import threading
from threading import Thread
from collections import OrderedDict

from sgtn_properties import Properties
from sgtn_util import FileUtil, NetUtil, SysUtil
from sgtn_util import LOG_TYPE_INFO, KEY_RESULT, KEY_HEADERS, KEY_ERROR
from sgtn_bykey import SingletonByKey
from sgtn_locale import SingletonLocale, SingletonLocaleUtil

from I18N import Config, Release, Translation

KEY_LOCALE = 'locale'
KEY_SOURCE = 'source'
KEY_ITEMS = 'format_items'

KEY_RESPONSE = 'response'
KEY_CODE = 'code'
KEY_DATA = 'data'
KEY_MESSAGES = 'messages'

KEY_PRODUCT = 'product'
KEY_VERSION = 'l10n_version'
KEY_SERVICE_URL = 'online_service_url'
KEY_OFFLINE_URL = 'offline_resources_base_url'
KEY_LOCAL_PATH = 'offline_resources_path'

KEY_DEFAULT_LOCALE = 'default_locale'
KEY_SOURCE_LOCALE = 'source_locale'
KEY_TRYDELAY = 'try_delay'
KEY_INTERVAL = 'cache_expired_time'
KEY_CACHEPATH = 'cache_path'
KEY_CACHETYPE = 'cache_type'
KEY_LOGPATH = 'log_path'
KEY_COMPONENTS = 'components'
KEY_LOCALES = 'locales'
KEY_LANG_TAG = 'language_tag'
KEY_COMPONENT_TAG = 'name'

KEY_COMPONENT_TEMPLATE = "component_template"
KEY_LOCALES_REFER = "locales_refer"
KEY_TEMPLATE = "template"

HEADER_REQUEST_ETAG = "If-None-Match"

LOCALE_DEFAULT = 'en-US'
MAX_THREAD = 1000
NOT_IMP_EXCEPTION = 'NotImplementedException'
LOCAL_TYPE_FILE = 'file'
LOCAL_TYPE_HTTP = 'http'
RES_TYPE_PROPERTIES = '.properties'
RES_TYPE_SGTN = '.json'


class ClientUtil:

    @classmethod
    def check_response_valid(cls, dict):
        if dict and KEY_RESULT in dict:
            status = dict[KEY_RESULT].get(KEY_RESPONSE)
            if status and KEY_CODE in status:
                code = status[KEY_CODE]
                if code == 200 or code == 604:
                    return True
        return False

    @classmethod
    def read_resource_files(cls, local_type, file_list):
        props = OrderedDict()

        try:
            for prop_file in file_list:
                if prop_file.endswith(RES_TYPE_PROPERTIES):
                    text = None
                    if (local_type == LOCAL_TYPE_HTTP):
                        text = NetUtil.http_get_text(prop_file)
                    else:
                        text = FileUtil.read_text_file(prop_file)
                    if text:
                        m = Properties().parse(text)
                        props.update(m)
                elif prop_file.endswith(RES_TYPE_SGTN):
                    m = None
                    if (local_type == LOCAL_TYPE_HTTP):
                        code, dt = NetUtil.http_get(prop_file, None)
                        if (code == 200):
                            m = dt.get(KEY_RESULT)
                    else:
                        m = FileUtil.read_json_file(prop_file)
                    if m:
                        m = m.get(KEY_MESSAGES)
                        props.update(m)
        except Exception as error:
            raise IOError('Error in loading property file. Check file(s) = ', file_list, ' ', error)

        return props


class SingletonConfig(Config):

    def __init__(self, base_path, config_data):
        self.base = base_path
        self.config_data = config_data

        self.product = config_data.get(KEY_PRODUCT)
        self.version = '%s' % config_data.get(KEY_VERSION)

        self.remote_url = config_data.get(KEY_SERVICE_URL)
        self.local_url = config_data.get(KEY_OFFLINE_URL)

        if self.local_url:
            parts = self.local_url.split('/')
            self.local_type = parts[0][:-1]

            if self.local_type == LOCAL_TYPE_FILE:
                start = 2
                needBasePath = False
                if len(parts) > 3:
                    if parts[3] == '..' or parts[3] == '.':
                        start = 3
                        needBasePath = True
                    if parts[3].endswith(':'):
                        start = 3
                self.local_url = '/'.join(parts[start:])
                if needBasePath:
                    self.local_url = os.path.join(base_path, self.local_url)

        self.log_path = self.get_path(KEY_LOGPATH)          # log path
        self.cache_path = self.get_path(KEY_CACHEPATH)      # cache path
        self.cache_type = self.get_item(KEY_CACHETYPE, 'default')   # cache type

        self.cache_expired_time = self.get_item(KEY_INTERVAL, 3600) # cache expired time
        self.try_delay = self.get_item(KEY_TRYDELAY, 10)    # try delay

        self.default_locale = self.get_item(KEY_DEFAULT_LOCALE, LOCALE_DEFAULT)
        self.source_locale = self.get_item(KEY_SOURCE_LOCALE, self.default_locale)

        self._expand_components()

    def _expand_locales(self, locales_def_array, template):
        locales = {}
        for k in range(len(locales_def_array)):
            locale_def = copy.deepcopy(locales_def_array[k])
            locales[locale_def.get(KEY_LANG_TAG)] = locale_def
            if KEY_LOCAL_PATH not in locale_def and template:
                locale_def[KEY_LOCAL_PATH] = copy.deepcopy(template.get(KEY_LOCAL_PATH))
        return locales

    def _expand_components(self):
        self.components = None
        components = self.config_data.get(KEY_COMPONENTS)
        if not components:
            return

        expand = {}
        self.components = {}
        for i in range(len(components)):
            component = components[i]
            if KEY_LOCALES in component:
                component[KEY_LOCALES] = self._expand_locales(component[KEY_LOCALES], None)
                self.components[component.get(KEY_COMPONENT_TAG)] = copy.deepcopy(component)
                continue

            template_name = component.get(KEY_TEMPLATE)
            if not template_name:
                template_name = KEY_COMPONENT_TEMPLATE

            if template_name not in expand:
                t = self.config_data.get(template_name)
                refer_name = t.get(KEY_LOCALES_REFER)

                refer = self.config_data.get(refer_name)
                if not refer:
                    continue
                expand[template_name] = self._expand_locales(refer, t)

            component[KEY_LOCALES] = expand[template_name]
            self.components[component.get(KEY_COMPONENT_TAG)] = copy.deepcopy(component)

    def get_config_data(self):
        # method of Config
        return self.config_data

    def get_info(self):
        # method of Config
        info = {'product': self.product, 'version': self.version,
                'remote': self.remote_url, 'local': self.local_url,
                'source_locale': self.source_locale, 'default_locale': self.default_locale}
        return info

    def extract_list(self, key, key_name, key_refer, refer):
        _dict = {}
        _define = self.config_data.get(key)
        if not _define:
            return None
        for one in _define:
            dup = copy.deepcopy(one)
            del dup[key_name]
            _dict[one[key_name]] = dup
            if key_refer not in dup and refer:
                dup[key_refer] = copy.deepcopy(refer)
        return _dict

    def get_item(self, key, default_value):
        value = self.config_data.get(key)
        if value is None:
            value = default_value
        return value

    def get_path(self, key):
        path = self.config_data.get(key)
        if path:
            if path.startswith('./') or path.startswith('../'):
                path = os.path.realpath(os.path.join(self.base, path))
        return path


class SingletonApi:
    VIP_PATH_HEAD = '/i18n/api/v2/translation/products/%s/versions/%s/'
    VIP_PARAMETER = 'pseudo=false&machineTranslation=false&checkTranslationStatus=false'
    VIP_GET_COMPONENT = 'locales/%s/components/%s?'

    def __init__(self, release_obj):
        self.rel = release_obj
        self.cfg = release_obj.cfg
        self.addr = self.cfg.remote_url

    def get_component_api(self, component, locale):
        head = self.VIP_PATH_HEAD % (self.cfg.product, self.cfg.version)
        path = self.VIP_GET_COMPONENT % (locale, component)
        return '%s%s%s%s' % (self.addr, head, path, self.VIP_PARAMETER)

    def get_localelist_api(self):
        head = self.VIP_PATH_HEAD % (self.cfg.product, self.cfg.version)
        return '%s%slocalelist' % (self.addr, head)

    def get_componentlist_api(self):
        head = self.VIP_PATH_HEAD % (self.cfg.product, self.cfg.version)
        return '%s%scomponentlist' % (self.addr, head)


class SingletonUpdateThread(Thread):

    def __init__(self, obj):
        Thread.__init__(self)
        self.obj = obj

    def run(self):
        self.obj.get_from_remote()


class SingletonAccessRemoteTask:

    def __init__(self, release_obj, obj):
        self.rel = release_obj
        self.obj = obj

        self.last_time = 0
        self.querying = False
        self.interval = self.rel.interval

    def set_retry(self, current):
        # try again after try_delay seconds
        self.last_time = current - self.interval + self.rel.try_delay

    def check(self):
        if not self.rel.cfg.remote_url:
            return

        access_remote = False
        if self.interval > 0:
            current = time.time()
            if current > self.last_time + self.interval:
                access_remote = True
        else:
            if self.last_time == 0:
                access_remote = True

        if not access_remote:
            return

        if self.querying:
            if self.obj.get_data_count() == 0:
                while(self.querying):
                    time.sleep(0.1)
            return

        self.querying = True
        if self.obj.get_data_count() == 0:
            self.obj.get_from_remote()
        else:
            th = SingletonUpdateThread(self.obj)
            th.start()


class SingletonComponent:

    def __init__(self, release_obj, locale, component, isLocalSource):
        self.rel = release_obj
        self.locale = locale
        self.localeItem = self.rel.bykey.get_locale_item(locale, isLocalSource)
        self.componentIndex = self.rel.bykey.get_component_index(component)
        self.component = component
        self.isLocalSource = isLocalSource
        self.countOfMessages = 0
        self.etag = None
        self.cache_path = None
        self.task = None if isLocalSource else SingletonAccessRemoteTask(release_obj, self)

        if self.task and self.rel.cache_path:
            self.cache_path = os.path.join(self.rel.cache_path, component, 'messages_%s.json' % locale)
            self.rel.log('--- cache file --- %s ---' % self.cache_path)

            if os.path.exists(self.cache_path):
                dt = FileUtil.read_json_file(self.cache_path)
                if KEY_MESSAGES in dt:
                    self.task.last_time = os.path.getmtime(self.cache_path)
                    self.set_messages(dt[KEY_MESSAGES])

    def set_messages(self, messages):
        for key in messages:
            text = messages[key]
            self.rel.bykey.set_string(key, self, self.componentIndex, self.localeItem, text)
        self.countOfMessages = len(messages)

    def get_messages(self):
        return self.rel.bykey.get_messages(self.componentIndex, self.localeItem)

    def get_message(self, key):
        return self.rel.bykey.get_string(key, self.componentIndex, self.localeItem)

    def is_messages_same(self, messages):
        for key in messages:
            text = messages[key]
            message = self.rel.bykey.get_string(key, self.componentIndex, self.localeItem)
            if message != text:
                return False
        return True

    def get_from_remote(self):
        current = time.time()

        try:
            # get messages
            addr = self.rel.api.get_component_api(self.component, self.locale)
            headers = {}
            if self.etag:
                headers[HEADER_REQUEST_ETAG] = self.etag
            code, dt = NetUtil.http_get(addr, headers)
            if code == 200 and ClientUtil.check_response_valid(dt):
                self.etag, interval = NetUtil.get_etag_maxage(dt.get(KEY_HEADERS))
                if interval:
                    self.task.interval = interval

                messages = dt[KEY_RESULT][KEY_DATA][KEY_MESSAGES]
                if self.cache_path:
                    if os.path.exists(self.cache_path) and self.is_messages_same(messages):
                        os.utime(self.cache_path, (current, current))
                    else:
                        self.rel.log('--- save --- %s ---' % self.cache_path)
                        FileUtil.save_json_file(self.cache_path, dt[KEY_RESULT][KEY_DATA])
                self.set_messages(messages)
                self.task.last_time = current
            elif code == 304:
                self.task.last_time = current
            else:
                self.task.set_retry(current)
        except Exception as e:
            self.task.set_retry(current)

        self.task.querying = False

    def get_data_count(self):
        return self.countOfMessages


class SingletonUseLocale:

    def __init__(self, singletonLocale, sourceLocale, isLocalSource, bykey):
        self.singletonLocale = singletonLocale
        self.locale = self.singletonLocale.get_original_locale()
        self.isLocalSource = isLocalSource

        singletonSourceLocale = SingletonLocaleUtil.get_singleton_locale(sourceLocale)
        self.isSourceLocale = self.locale in singletonSourceLocale.get_near_locale_list()

        self.localeItem = bykey.get_locale_item(self.locale, True) if isLocalSource and bykey else None

        self.components = {}


class SingletonReleaseBase:

    def __init__(self, cfg):
        self.cfg = cfg
        self.cache_path = None
        self.scope = None
        self.logger = None
        self.interval = 0
        self.try_delay = 0
        self.detach = False

        self.locale_list = []
        self.component_list = []

        self.remote_pool = {}
        self.source_pool = {}
        self.local_handled = {}
        self.component_handled = {}

        if not cfg:
            return

        if cfg.log_path:
            log_file = os.path.join(cfg.log_path, '%s_%s.log' % (self.cfg.product, self.cfg.version))
            self.init_logger(log_file)

        if cfg.cache_path:
            self.cache_path = os.path.join(cfg.cache_path, self.cfg.product, self.cfg.version)
            self.log('--- cache path --- %s ---' % self.cache_path)

        self.interval = cfg.cache_expired_time
        self.try_delay = cfg.try_delay

        self.task = SingletonAccessRemoteTask(self, self)
        self.get_scope()

        self.remote_default_locale = self.get_locale_supported(self.cfg.default_locale)
        self.remote_source_locale = self.get_locale_supported(self.cfg.source_locale)

        self.isDifferent = self.remote_default_locale != self.remote_source_locale

        self.bykey = SingletonByKey(self.cfg.source_locale, self.cfg.default_locale, self.isDifferent, self.cfg.cache_type)

        self.useSourceLocale = self.get_use_locale(self.cfg.source_locale, True)
        self._get_local_resource(self.useSourceLocale, self.cfg.source_locale)

        self.useDefaultLocale = None
        if self.isDifferent:
            self.useDefaultLocale = self.get_use_locale(self.cfg.default_locale, False)


    def get_use_locale(self, locale, asSource):
        pool = self.source_pool if asSource else self.remote_pool
        useLocale = pool.get(locale)
        if useLocale is None:
            singletonLocale = SingletonLocaleUtil.get_singleton_locale(locale)
            useLocale = singletonLocale.find_item(pool, 1)

            if useLocale is None:
                useLocale = SingletonUseLocale(singletonLocale, self.cfg.source_locale, asSource, self.bykey)

            for one in useLocale.singletonLocale.get_near_locale_list():
                if one not in pool:
                    pool[one] = useLocale
        return useLocale

    def get_scope(self):
        self.api = SingletonApi(self)

        if self.cache_path:
            self.locale_list = FileUtil.read_json_file(os.path.join(self.cache_path, 'locale_list.json'))
            self.component_list = FileUtil.read_json_file(os.path.join(self.cache_path, 'component_list.json'))
        if not self.cfg.remote_url:
            return

        if not self.locale_list:
            self.get_from_remote()
        else:
            th = SingletonUpdateThread(self)
            th.start()

    def get_from_remote(self):
        self.task.last_time = time.time()

        try:
            # get locale list
            scope = self._get_scope_item(self.api.get_localelist_api(), KEY_LOCALES, 'locale_list.json')
            if scope:
                self.locale_list = scope

            # get component list
            scope = self._get_scope_item(self.api.get_componentlist_api(), KEY_COMPONENTS, 'component_list.json')
            if scope:
                self.component_list = scope
        except Exception as e:
            pass

        self.task.querying = False

    def get_data_count(self):
        if not self.locale_list or not self.component_list:
            return 0
        return len(self.locale_list) + len(self.component_list)

    def init_logger(self, log_file):
        self.logger = SysUtil.init_logger(log_file, 'sgtn_%s_%s' % (self.cfg.product, self.cfg.version))
        self.log('--- release --- %s --- %s --- %s ---' % (self.cfg.product, self.cfg.version, time.time()))

    def log(self, text, log_type = LOG_TYPE_INFO):
        SysUtil.log(self.logger, text, log_type)

    def _load_one_local(self, component, locale, path_define):
        if not path_define:
            return None
        for i in range(len(path_define)):
            path = path_define[i].replace('$COMPONENT', component).replace('$LOCALE', locale)
            path_define[i] = os.path.join(self.cfg.local_url, path)
        return ClientUtil.read_resource_files(self.cfg.local_type, path_define)

    def _get_scope_item(self, addr, key, keep_name):
        code, dt = NetUtil.http_get(addr, None)
        if code == 200 and ClientUtil.check_response_valid(dt):
            _, interval = NetUtil.get_etag_maxage(dt.get(KEY_HEADERS))
            if interval:
                self.task.interval = interval

            scope = dt[KEY_RESULT][KEY_DATA][key]
            if scope and self.cache_path:
                FileUtil.save_json_file(os.path.join(self.cache_path, keep_name), scope)
            return scope
        return None

    def _extract_info_from_dir(self, root):
        if self.cfg.local_type != LOCAL_TYPE_FILE:
            return

        components = {}
        dir_list, _ = FileUtil.get_dir_info(root)
        for component in dir_list:
            components[component] = {}
            component_obj = components[component]
            component_obj[KEY_LOCALES] = {}
            locales_cfg = component_obj.get(KEY_LOCALES)

            component_path = os.path.join(self.cfg.local_url, component)
            _, file_list = FileUtil.get_dir_info(component_path)
            for res_file in file_list:
                parts = re.split(r"messages(.*)\.", res_file)
                if (len(parts) == 3):
                    if parts[1].startswith('_'):
                        locale = parts[1][1:]
                    elif parts[1] == '':
                        locale = self.cfg.source_locale
                    locales_cfg[locale] = {KEY_LOCAL_PATH: [os.path.join(component, res_file)]}
        return components

    def _get_local_resource(self, useLocale, locale):
        if useLocale is None:
            return
        locale_item = useLocale.components

        if not self.cfg.local_url:
            return

        if not self.cfg.components:
            self.cfg.components = self._extract_info_from_dir(self.cfg.local_url)
            if not self.cfg.components:
                return

        for component in self.cfg.components:
            locales_cfg = self.cfg.components[component].get(KEY_LOCALES)
            locale_define = None
            if locales_cfg:
                singletonLocale = SingletonLocaleUtil.get_singleton_locale(locale)
                locale_define = singletonLocale.find_item(locales_cfg, 0)

            combineKey = locale + '_!_' + component
            if locale_define and combineKey not in self.local_handled:
                path_define = locale_define.get(KEY_LOCAL_PATH)
                map = self._load_one_local(component, locale, path_define)
                component_obj = SingletonComponent(self, locale, component, useLocale.isLocalSource)
                component_obj.set_messages(map)
                locale_item[component] = component_obj
            self.local_handled[combineKey] = True

    def _get_remote_resource(self, locale, component):
        if not self.locale_list or not self.component_list:
            return None
        singletonLocale = SingletonLocaleUtil.get_singleton_locale(locale)
        if not singletonLocale.is_in_locale_list(self.locale_list):
            return None
        if component not in self.component_list:
            return None

        components = self.get_use_locale(locale, False).components
        component_obj = components.get(component)
        if component_obj is None:
            self.log('--- component --- %s ---' % component)
            component_obj = SingletonComponent(self, locale, component, False)
            components[component] = component_obj

        component_obj.task.check()
        return component_obj

    def _get_component(self, locale, component):
        component_remote = self._get_remote_resource(locale, component)
        if component_remote:
            return component_remote

        component_obj = None
        if self.cfg.local_url:
            useLocale = self.get_use_locale(locale, False)
            combineKey = locale + '_!_' + component
            if combineKey not in self.local_handled:
                self._get_local_resource(useLocale, locale)
                self.local_handled[combineKey] = True
            if useLocale:
                component_obj = useLocale.components.get(component)
                if component_obj is None and useLocale.isSourceLocale:
                    component_obj = self.useSourceLocale.components.get(component)

        return component_obj

    def _get_message(self, component, key, source, locale):
        message = source if source != None else key
        if not key or not locale:
            return message

        if not self.bykey._onlyByKey and not component:
            return message

        self.task.check()

        componentIndex = self.bykey.get_component_index(component)
        if componentIndex >= 0:
            combineKey = locale + '_!_' + component
            if combineKey not in self.component_handled:
                self._get_component(self.remote_source_locale, component)
                componentObj = self._get_component(locale, component)
                if self.isDifferent:
                    self._get_component(self.remote_default_locale, component)
                if componentObj:
                    self.component_handled[combineKey] = True

        localeItem = self.bykey.get_locale_item(locale, False)
        message = self.bykey.get_string(key, componentIndex, localeItem, True)
        return message


class SingletonRelease(SingletonReleaseBase, Release, Translation):

    def get_config(self):
        # method of Release
        return self.cfg

    def get_translation(self):
        # method of Release
        return self

    def get_locale_strings(self, locale, asSource):
        # method of Translation
        collect = {}
        useLocale = self.get_use_locale(locale, asSource)
        if useLocale and useLocale.components:
            components = useLocale.components
            for component in components:
                collect[component] = components[component].get_messages()
        return collect

    def get_source(self, component, key, sourceInCode):
        componentIndex = self.bykey.get_component_index(component)
        source = self.bykey.get_string(key, componentIndex, self.useSourceLocale.localeItem, False)
        if source != None:
            return source

        source = self._get_message(component, key, sourceInCode, self.cfg.source_locale)
        return source

    def get_raw(self, component, key, sourceInCode, locale, items):
        useLocale = self.get_use_locale(locale, False)
        if useLocale.isSourceLocale:
            if sourceInCode != None:
                return sourceInCode
            return self.get_source(component, key, sourceInCode)

        source = self.get_source(component, key, sourceInCode)
        if sourceInCode != None and source != None and source != sourceInCode:
            return sourceInCode

        return self._get_message(component, key, source, locale)

    def get_string(self, component, key, **kwargs):
        # method of Translation
        sourceInCode = kwargs.get(KEY_SOURCE) if kwargs else None
        locale = kwargs.get(KEY_LOCALE) if kwargs else None
        items = kwargs.get(KEY_ITEMS) if kwargs else None

        if not locale:
            locale = SingletonClientManager().get_current_locale()
            
        text = self.get_raw(component, key, sourceInCode, locale, items)
        if text and items:
            if isinstance(items, list):
                text = text.format(*items)
            elif isinstance(items, dict):
                text = text.format(**items)

        if text is None:
            text = key
        return text

    def get_locale_supported(self, locale):
        # method of Translation
        return SysUtil.get_fallback_locale(locale)


class SingletonClientManager(object):
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
            cls._instance.init()
        return cls._instance

    def init(self):
        self._products = {}

    def add_config_file(self, config_file, replaceMap = None):
        config_text = FileUtil.read_text_file(config_file)
        if replaceMap:
            for key in replaceMap:
                config_text = config_text.replace(key, replaceMap[key])
        config_data = FileUtil.parse_datatree(config_text)

        base_path = os.path.dirname(os.path.realpath(config_file))
        cfg = self.add_config(base_path, config_data)
        return cfg

    def add_config(self, base_path, config_data):
        if not config_data:
            return
        cfg = SingletonConfig(base_path, config_data)
        release_obj = self.get_release(cfg.product, cfg.version)
        if release_obj is None:
            self.create_release(cfg)
        return cfg

    def get_release(self, product, version):
        if not product or not version:
            return None

        releases = self._products.get(product)
        if releases is None:
            return None

        return releases.get(version)

    def create_release(self, cfg):
        if not cfg or not cfg.product or not cfg.version:
            return

        releases = self._products.get(cfg.product)
        if releases is None:
            self._products[cfg.product] = {}
            releases = self._products.get(cfg.product)

        release_obj = releases.get(cfg.version)
        if release_obj is None:
            release_obj = SingletonRelease(cfg)
            releases[cfg.version] = release_obj

    def set_current_locale(self, locale):
        current = sys._getframe().f_back.f_back
        for i in range(10):
            if not hasattr(current, 'f_locals'):
                break

            locals = current.f_locals
            locals['_singleton_locale_'] = locale

            if not hasattr(current, 'f_back'):
                break
            current = current.f_back

    def get_current_locale(self):
        current = sys._getframe().f_back.f_back
        for i in range(10):
            if not hasattr(current, 'f_locals'):
                break

            locals = current.f_locals
            if '_singleton_locale_' in locals:
                return locals['_singleton_locale_']

            if not hasattr(current, 'f_back'):
                break
            current = current.f_back
        return LOCALE_DEFAULT



# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#

import sys
import unittest

sys.path.append('../sgtnclient')
if sys.version_info.major == 2:
    # Support utf8 text
    reload(sys)
    sys.setdefaultencoding('utf-8')

from sgtn_locale import SingletonLocale, SingletonLocaleUtil
from sgtn_bykey import SingletonByKey, SingletonByKeyLocale, SingletonByKeyComponents


class TestByKey(unittest.TestCase):

    def check_one(self, bykey, key, message, idComponent, localeObj):
        bykey.set_string(key, None, idComponent, localeObj, message)
        msg = bykey.get_string(key, -1, localeObj)
        print('--- message --- %s --- %s' % (key, msg))

    def show_one(self, bykey, key, localeObj):
        msg = bykey.get_string(key, -1, localeObj)
        print('--- message --- %s --- %s' % (key, msg))

    def test_release(self):

        bykey = SingletonByKey('en', 'en', False, 'by_key')
        id1 = bykey.get_component_index('first')
        id2 = bykey.get_component_index('second')
        print('--- component id --- %s' % [id1, id2])

        localeObj1 = bykey.get_locale_item('en', True)
        localeObj2 = bykey.get_locale_item('en', False)
        localeObj3 = bykey.get_locale_item('de_de', False)
        localeObj4 = bykey.get_locale_item('de-DE', False)

        item = bykey.get_key_item(0, 0)

        bykey.set_string("key1", None, id1, localeObj1, "en_message")
        bykey.set_string("key1", None, id1, localeObj2, "en_message")

        self.check_one(bykey, "key1", "de_message", id1, localeObj3)
        self.check_one(bykey, "key2", "de_message3", id1, localeObj3)
        self.check_one(bykey, "key1", "de_message2", id1, localeObj3)

        self.show_one(bykey, "key1", localeObj4)

        messages = bykey.get_messages(id1, localeObj3)
        print('--- messages --- %s' % messages)

    def test_locale(self):
        sgtnLocale = SingletonLocaleUtil.get_singleton_locale(None)
        print('--- locale --- %s' % sgtnLocale.get_near_locale_list())

        sgtnLocale = SingletonLocaleUtil.get_singleton_locale('zh_cn')
        print('--- locale --- %s' % sgtnLocale.get_near_locale_list())

        sgtnLocale = SingletonLocaleUtil.get_singleton_locale('de_de')
        print('--- locale --- %s' % sgtnLocale.get_near_locale_list())


if __name__ == '__main__':
    unittest.main()




# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#

import sys
import unittest
import time
import threading
from threading import Thread

sys.path.append('../sgtnclient')
if sys.version_info.major == 2:
    # Support utf8 text
    reload(sys)
    sys.setdefaultencoding('utf-8')

import I18N

count = 0
WAIT = 6

from util import logger, Util


class WorkThread(Thread):

    def __init__(self, locale):
        global count
        count += 1
        Thread.__init__(self)
        self.locale = locale
        self.count = count

    def run(self):
        global logger

        I18N.set_current_locale(self.locale)

        time.sleep(WAIT)

        theLocale = I18N.get_current_locale()

        thid = threading.current_thread().ident
        logger.info('--- %s --- th%s --- %s ---' % (self.count, thid, theLocale))


class TestMultiTask(unittest.TestCase):

    def do_test_current_locale_in_threads(self):
        global logger

        locales = ['en', 'en-US', 'de', 'zh-CN', 'fr']

        for k in range(3):
            for i in range(10000):
                th = WorkThread(locales[i % len(locales)])
                th.start()
            time.sleep(1)
            logger.info('>\n>\n>\n')

        I18N.set_current_locale('de')
        theLocale = I18N.get_current_locale()
        time.sleep(WAIT)

    def do_test_current_locale_in_async(self):
        global logger
        if Util.is_async_supported():
            from async_util import AsyncWork 
            AsyncWork().hello()

    def test_all(self):
        self.do_test_current_locale_in_threads()
        self.do_test_current_locale_in_async()
        
        logger.info('--- end ---')
        time.sleep(1)


if __name__ == '__main__':
    unittest.main()




# -*-coding:UTF-8 -*-
#
# Copyright 2020-2021 VMware, Inc.
# SPDX-License-Identifier: EPL-2.0
#

from setuptools import setup, find_packages

setup(
    name="sgtnclient-test",
    version="0.6.0.1",
    author="Jasper Jin",
    author_email="jasperj@vmware.com",
    description="Singleton Client",
    license="EPL-2.0",
    url="https://github.com/vmware/singleton/tree/g11n-python-client",
    packages=['sgtnclient'],
    install_requires=[
        "PyYAML"
        ],
    classifiers=[
        "Operating System :: OS Independent"
    ]
)
